@@ -3,7 +3,6 @@ from typing import List
 
 from .LcleanCommand import LcleanCommand
 from .LstartCommand import LstartCommand
-from ... import utils
 from ...foundation.cli.command.Command import Command
 from ...strings import strings, wiki_description
 
@@ -23,7 +22,7 @@ class LrestartCommand(Command):
             '-h', '--help',
             action='help',
             default=argparse.SUPPRESS,
-            help='Show an help message and exit.'
+            help='Show a help message and exit.'
         )
 
         group = self.parser.add_mutually_exclusive_group(required=False)
@@ -43,7 +42,7 @@ class LrestartCommand(Command):
             const=True,
             help='Start the network scenario opening terminal windows.'
         )
-        group.add_argument(
+        self.parser.add_argument(
             "--privileged",
             action="store_const",
             const=True,
@@ -70,11 +69,11 @@ class LrestartCommand(Command):
         )
         self.parser.add_argument(
             '-o', '--pass',
-            dest='options',
-            metavar="OPTION",
+            dest='global_machine_metadata',
+            metavar="METADATA",
             nargs='*',
             required=False,
-            help="Apply options to all devices of a network scenario during startup."
+            help="Apply metadata to all devices of a network scenario during startup."
         )
         self.parser.add_argument(
             '--xterm',
@@ -111,6 +110,14 @@ class LrestartCommand(Command):
             const=True,
             help='Mount "/shared" directory inside devices.'
         )
+        self.parser.add_argument(
+            '--exclude',
+            dest='excluded_machines',
+            metavar='DEVICE_NAME',
+            nargs='+',
+            default=[],
+            help='Exclude specified devices.'
+        )
         self.parser.add_argument(
             'machine_name',
             metavar='DEVICE_NAME',
@@ -118,17 +125,18 @@ class LrestartCommand(Command):
             help='Restarts only specified devices.'
         )
 
-    def run(self, current_path: str, argv: List[str]) -> None:
+    def run(self, current_path: str, argv: List[str]) -> int:
         self.parse_args(argv)
         args = self.get_args()
 
-        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
-        lab_path = utils.get_absolute_path(lab_path)
-
         lclean_argv = ['-d', args['directory']] if args['directory'] else []
 
         if args['machine_name']:
             lclean_argv.extend(args['machine_name'])
+        if args['excluded_machines']:
+            lclean_argv.extend(['--exclude'] + args['excluded_machines'])
 
-        LcleanCommand().run(lab_path, lclean_argv)
-        LstartCommand().run(lab_path, argv)
+        LcleanCommand().run(current_path, lclean_argv)
+        LstartCommand().run(current_path, argv)
+
+        return 0
