@@ -1,16 +1,13 @@
 import argparse
-import logging
 import os
-import sys
 from typing import List
 
-from ..ui.utils import create_table
-from ..ui.utils import format_headers
+from ..ui.utils import create_lab_table
+from ..ui.utils import create_panel, LabMetaHighlighter
 from ... import utils
 from ...exceptions import PrivilegeError, EmptyLabError
 from ...foundation.cli.command.Command import Command
 from ...manager.Kathara import Kathara
-from ...model.Lab import Lab
 from ...parser.netkit.DepParser import DepParser
 from ...parser.netkit.ExtParser import ExtParser
 from ...parser.netkit.FolderParser import FolderParser
@@ -35,7 +32,7 @@ class LstartCommand(Command):
             '-h', '--help',
             action='help',
             default=argparse.SUPPRESS,
-            help='Show an help message and exit.'
+            help='Show a help message and exit.'
         )
 
         group = self.parser.add_mutually_exclusive_group(required=False)
@@ -55,7 +52,7 @@ class LstartCommand(Command):
             const=True,
             help='Start the network scenario opening terminal windows.'
         )
-        group.add_argument(
+        self.parser.add_argument(
             "--privileged",
             action="store_const",
             const=True,
@@ -82,19 +79,19 @@ class LstartCommand(Command):
         )
         self.parser.add_argument(
             '-o', '--pass',
-            dest='options',
-            metavar="OPTION",
+            dest='global_machine_metadata',
+            metavar="METADATA",
             nargs='*',
             required=False,
-            help="Apply options to all devices of a network scenario during startup."
+            help="Apply metadata to all devices of a network scenario during startup."
         )
         self.parser.add_argument(
-            '--xterm',
+            '--xterm', '--terminal-emu',
             required=False,
             help='Set a different terminal emulator application (Unix only).'
         )
         self.parser.add_argument(
-            '--print',
+            '--print', '--dry-mode',
             dest="dry_mode",
             required=False,
             action='store_true',
@@ -130,6 +127,14 @@ class LstartCommand(Command):
             const=True,
             help='Mount "/shared" directory inside devices.'
         )
+        self.parser.add_argument(
+            '--exclude',
+            dest='excluded_machines',
+            metavar='DEVICE_NAME',
+            nargs='+',
+            default=[],
+            help='Exclude specified devices from startup.'
+        )
         self.parser.add_argument(
             'machine_name',
             metavar='DEVICE_NAME',
@@ -137,28 +142,26 @@ class LstartCommand(Command):
             help='Launches only specified devices.'
         )
 
-    def run(self, current_path: str, argv: List[str]) -> Lab:
+    def run(self, current_path: str, argv: List[str]) -> int:
         self.parse_args(argv)
         args = self.get_args()
 
         lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
         lab_path = utils.get_absolute_path(lab_path)
 
+        # Load custom 'kathara.conf' if it exists
+        self._load_custom_configuration(lab_path)
+
         Setting.get_instance().open_terminals = args['terminals'] if args['terminals'] is not None \
             else Setting.get_instance().open_terminals
         Setting.get_instance().terminal = args['xterm'] or Setting.get_instance().terminal
 
-        if args['privileged']:
-            if not utils.is_admin():
-                raise PrivilegeError("You must be root in order to start Kathara devices in privileged mode.")
-            else:
-                logging.warning("Running devices with privileged capabilities, terminals won't open!")
-                Setting.get_instance().open_terminals = False
-
-        if args['dry_mode']:
-            logging.info(format_headers("Checking Network Scenario"))
-        else:
-            logging.info(format_headers("Starting Network Scenario"))
+        self.console.print(
+            create_panel(
+                "Checking Network Scenario" if args['dry_mode'] else "Starting Network Scenario",
+                style="blue bold", justify="center"
+            )
+        )
 
         try:
             lab = LabParser.parse(lab_path)
@@ -174,22 +177,19 @@ class LstartCommand(Command):
             lab.apply_dependencies(dependencies)
 
         lab_meta_information = str(lab)
-
         if lab_meta_information:
-            logging.info("\n" + lab_meta_information)
-            logging.info(format_headers())
+            meta_highlighter = LabMetaHighlighter()
+            self.console.print(create_panel(meta_highlighter(lab_meta_information)))
 
         if len(lab.machines) <= 0:
             raise EmptyLabError()
 
-        try:
-            options = OptionParser.parse(args['options'])
-            lab.general_options = {**lab.general_options, **options}
-        except ValueError as e:
-            raise e
+        lab.global_machine_metadata = OptionParser.parse(args['global_machine_metadata'])
 
         lab_ext_path = os.path.join(lab_path, 'lab.ext')
+        lab_ext_exists = False
         if os.path.exists(lab_ext_path):
+            lab_ext_exists = True
             if utils.is_platform(utils.LINUX) or utils.is_platform(utils.LINUX2):
                 if utils.is_admin():
                     external_links = ExtParser.parse(lab_path)
@@ -202,28 +202,41 @@ class LstartCommand(Command):
                 else:
                     raise PrivilegeError("You must be root in order to use lab.ext file.")
             else:
-                raise OSError("lab.ext is only available on UNIX systems.")
+                raise OSError("lab.ext is only available on Linux systems.")
 
         # If dry mode, we just check if the lab.conf is correct.
         if args['dry_mode']:
-            logging.info("lab.conf file is correct.")
+            self.console.print("[green]\u2713 [bold]lab.conf[/bold] file is correct.")
             if dependencies:
-                logging.info("lab.dep file is correct.")
-            if os.path.exists(lab_ext_path):
-                logging.info("lab.ext file is correct.")
+                self.console.print("[green]\u2713 [bold]lab.dep[/bold] file is correct.")
+            if lab_ext_exists:
+                self.console.print("[green]\u2713 [bold]lab.ext[/bold] file is correct.")
 
-            logging.info("Exiting...")
-
-            sys.exit(0)
+            return 0
 
         lab.add_option('hosthome_mount', args['hosthome_mount'])
         lab.add_option('shared_mount', args['shared_mount'])
-        lab.add_option('privileged_machines', args['privileged'])
 
-        Kathara.get_instance().deploy_lab(lab, selected_machines=set(args['machine_name']))
+        if args['privileged'] or any([x.is_privileged() for x in lab.machines.values()]):
+            if not utils.is_admin():
+                raise PrivilegeError("You must be root in order to start Kathara devices in privileged mode.")
+            else:
+                if Setting.get_instance().open_terminals:
+                    self.console.print(
+                        "[yellow]\u26a0 Running devices with privileged capabilities, terminals might not open!"
+                    )
+        lab.add_global_machine_metadata('privileged', args['privileged'])
+
+        Kathara.get_instance().deploy_lab(
+            lab, selected_machines=set(args['machine_name']), excluded_machines=set(args['excluded_machines'])
+        )
 
         if args['list']:
-            machines_stats = Kathara.get_instance().get_machines_stats(lab_hash=lab.hash)
-            print(next(create_table(machines_stats)))
+            with self.console.status(
+                    f"Loading...",
+                    spinner="dots"
+            ) as _:
+                machines_stats = Kathara.get_instance().get_machines_stats(lab_hash=lab.hash)
+                self.console.print(create_lab_table(machines_stats))
 
-        return lab
+        return 0
