@@ -5,9 +5,11 @@ from ...foundation.cli.command.Command import Command
 
 
 class SettingsCommand(Command):
-    def run(self, current_path: str, argv: List[str]) -> None:
+    def run(self, current_path: str, argv: List[str]) -> int:
         menu_factory = SettingsMenuFactory()
 
         menu = menu_factory.create_menu()
 
         menu.show()
+
+        return 0
