@@ -1,7 +1,7 @@
 import argparse
-import logging
 from typing import List
 
+from ..ui.utils import create_panel
 from ...foundation.cli.command.Command import Command
 from ...manager.Kathara import Kathara
 from ...model.Lab import Lab
@@ -23,21 +23,26 @@ class VcleanCommand(Command):
             '-h', '--help',
             action='help',
             default=argparse.SUPPRESS,
-            help='Show an help message and exit.'
+            help='Show a help message and exit.'
         )
 
         self.parser.add_argument(
             '-n', '--name',
             metavar='DEVICE_NAME',
             required=True,
-            help='Name of the device to be cleaned.'
+            help='The name of the device to clean.'
         )
 
-    def run(self, current_path: str, argv: List[str]) -> None:
+    def run(self, current_path: str, argv: List[str]) -> int:
         self.parse_args(argv)
         args = self.get_args()
 
         lab = Lab("kathara_vlab")
+
+        self.console.print(
+            create_panel(f"Stopping Device `{args['name']}`", style="blue bold", justify="center")
+        )
+
         Kathara.get_instance().undeploy_lab(lab_name=lab.name, selected_machines={args['name']})
 
-        logging.info("Device `%s` deleted successfully!" % args['name'])
+        return 0
