@@ -1,9 +1,7 @@
 import argparse
-import logging
-import sys
 from typing import List
 
-from ..ui.utils import format_headers, colon_separated
+from ..ui.utils import create_panel, interface_cd_mac, volume
 from ... import utils
 from ...exceptions import PrivilegeError
 from ...foundation.cli.command.Command import Command
@@ -28,7 +26,7 @@ class VstartCommand(Command):
             '-h', '--help',
             action='help',
             default=argparse.SUPPRESS,
-            help='Show an help message and exit.'
+            help='Show a help message and exit.'
         )
 
         group = self.parser.add_mutually_exclusive_group(required=False)
@@ -49,18 +47,18 @@ class VstartCommand(Command):
             help='Start the device opening its terminal window.'
         )
         group.add_argument(
+            '--num_terms',
+            metavar='NUM_TERMS',
+            required=False,
+            help='Choose the number of terminals to open for the device.'
+        )
+        self.parser.add_argument(
             "--privileged",
             action="store_const",
             const=True,
             required=False,
             help='Start the device in privileged mode. MUST BE ROOT FOR THIS OPTION.'
         )
-        group.add_argument(
-            '--num_terms',
-            metavar='NUM_TERMS',
-            required=False,
-            help='Choose the number of terminals to open for the device.'
-        )
         self.parser.add_argument(
             '-n', '--name',
             metavar='DEVICE_NAME',
@@ -69,9 +67,9 @@ class VstartCommand(Command):
         )
         self.parser.add_argument(
             '--eth',
-            type=colon_separated,
+            type=interface_cd_mac,
             dest='eths',
-            metavar='N:CD',
+            metavar='N:CD/MAC',
             nargs='+',
             required=False,
             help='Set a specific interface on a collision domain.'
@@ -114,12 +112,12 @@ class VstartCommand(Command):
             help='Mount "/hosthome" directory inside the device.'
         )
         self.parser.add_argument(
-            '--xterm',
+            '--xterm', '--terminal-emu',
             required=False,
             help='Set a different terminal emulator application (Unix only).'
         )
         self.parser.add_argument(
-            '--print',
+            '--print', '--dry-run',
             dest='dry_mode',
             required=False,
             action='store_true',
@@ -155,50 +153,99 @@ class VstartCommand(Command):
             required=False,
             help='Set environment variable for the device.'
         )
+        self.parser.add_argument(
+            '--ulimit',
+            dest='ulimits',
+            metavar='KEY=SOFT[:HARD]',
+            nargs='+',
+            required=False,
+            help='Set ulimit for the device.'
+        )
+        self.parser.add_argument(
+            '--volume',
+            type=volume,
+            dest='volumes',
+            metavar='HOST|GUEST|[MODE]',
+            nargs='+',
+            required=False,
+            help='Specify a volume to mount.'
+        )
         self.parser.add_argument(
             '--shell',
             required=False,
             help='Set the shell (sh, bash, etc.) that should be used inside the device.'
         )
+        self.parser.add_argument(
+            '--entrypoint',
+            metavar='ENTRYPOINT',
+            required=False,
+            help='Specify the entrypoint command of the device.'
+        )
+        self.parser.add_argument(
+            "args",
+            metavar='ARG',
+            nargs=argparse.REMAINDER,
+            help='Specify extra arguments for the entrypoint command.'
+        )
 
-    def run(self, current_path: str, argv: List[str]) -> None:
+    def run(self, current_path: str, argv: List[str]) -> int:
         self.parse_args(argv)
         args = self.get_args()
 
+        name = args.pop('name')
+
+        self.console.print(
+            create_panel(
+                f"Checking Device `{name}`" if args['dry_mode'] else f"Starting Device `{name}`",
+                style="blue bold", justify="center"
+            )
+        )
+
         if args['dry_mode']:
-            logging.info("Device configuration is correct. Exiting...")
-            sys.exit(0)
-        else:
-            logging.info(format_headers("Starting Device"))
+            self.console.print(f"[green]\u2713 [bold]{name}[/bold] configuration is correct.")
+            return 0
 
         Setting.get_instance().open_terminals = args['terminals'] if args['terminals'] is not None \
             else Setting.get_instance().open_terminals
         Setting.get_instance().terminal = args['xterm'] or Setting.get_instance().terminal
         Setting.get_instance().device_shell = args['shell'] or Setting.get_instance().device_shell
 
+        lab = Lab("kathara_vlab")
+        lab.add_option('hosthome_mount', args['hosthome_mount'])
+        lab.add_option('shared_mount', False)
+
         if args['privileged']:
             if not utils.is_admin():
                 raise PrivilegeError("You must be root in order to start this Kathara device in privileged mode.")
             else:
-                logging.warning("Running device with privileged capabilities, terminal won't open!")
-                Setting.get_instance().open_terminals = False
+                if Setting.get_instance().open_terminals:
+                    self.console.print(
+                        "[yellow]\u26a0 Running devices with privileged capabilities, terminal might not open!"
+                    )
+        lab.add_global_machine_metadata('privileged', args['privileged'])
 
-        lab = Lab("kathara_vlab")
-        lab.add_option('hosthome_mount', args['hosthome_mount'])
-        lab.add_option('shared_mount', False)
-        lab.add_option('privileged_machines', args['privileged'])
+        if args['args'] and args['args'][0] == "--":
+            args['args'] = args['args'][1:]
 
-        name = args.pop('name')
+        volumes = args.pop('volumes')
+        eths = args.pop('eths')
 
         device = lab.get_or_new_machine(name, **args)
 
-        if args['eths']:
-            for iface_number, cd in args['eths']:
+        if eths:
+            for iface_number, cd, mac_address in eths:
                 try:
-                    lab.connect_machine_to_link(device.name, cd, machine_iface_number=int(iface_number))
+                    lab.connect_machine_to_link(device.name, cd,
+                                                machine_iface_number=int(iface_number),
+                                                mac_address=mac_address)
                 except ValueError:
-                    raise SyntaxError("Interface number in `--eth %s:%s` is not a number." % (iface_number, cd))
+                    s = f"{cd}/{mac_address}" if mac_address else f"{cd}"
+                    raise SyntaxError(f"Interface number in `--eth {iface_number}:{s}` is not a number.")
 
-        lab.check_integrity()
+        if volumes:
+            for v in volumes:
+                device.add_meta("volume", v)
 
         Kathara.get_instance().deploy_lab(lab)
+
+        return 0
