@@ -28,7 +28,7 @@ class SettingsNotFoundError(Exception):
 
 class DockerDaemonConnectionError(Exception):
     def __init__(self, message: str) -> None:
-        super().__init__(f"Cannot connect to Docker Daemon: {message}")
+        super().__init__(f"Cannot connect to Docker Daemon, this may indicate that it is not running. {message}")
 
 
 class NotSupportedError(Exception):
@@ -95,6 +95,11 @@ class MachineNotFoundError(Exception):
     pass
 
 
+class MachineNotRunningError(Exception):
+    def __init__(self, machine_name: str) -> None:
+        super().__init__(f"Device `{machine_name}` is not running.")
+
+
 class MachineNotReadyError(Exception):
     def __init__(self, machine_name: str) -> None:
         super().__init__(f"Device `{machine_name}` is not ready.")
@@ -111,6 +116,14 @@ class MachineBinaryError(Exception):
         return f"Binary `{self.binary}` not found in device `{self.machine_name}`."
 
 
+# Interface Exceptions
+class InterfaceMacAddressError(Exception):
+    def __init__(self, mac_address: str, interface_num: int, machine_name: str) -> None:
+        super().__init__(
+            f"MAC address {mac_address} on interface `{interface_num}` of device `{machine_name}` is invalid."
+        )
+
+
 # Link Exceptions
 class LinkNotFoundError(Exception):
     pass
