@@ -16,7 +16,7 @@ from .KubernetesNamespace import KubernetesNamespace
 from .stats.KubernetesLinkStats import KubernetesLinkStats
 from ... import utils
 from ...event.EventDispatcher import EventDispatcher
-from ...exceptions import LinkNotFoundError
+from ...exceptions import InvocationError
 from ...model.Lab import Lab
 from ...model.Link import Link
 from ...setting.Setting import Setting
@@ -39,23 +39,35 @@ class KubernetesLink(object):
 
         self.seed: str = KubernetesConfig.get_cluster_user()
 
-    def deploy_links(self, lab: Lab, selected_links: Set[str] = None) -> None:
+    def deploy_links(self, lab: Lab, selected_links: Set[str] = None, excluded_links: Set[str] = None) -> None:
         """Deploy all the links contained in lab.links.
 
         Args:
             lab (Kathara.model.Lab.Lab): A Kathara network scenario.
             selected_links (Set[str]): A set containing the name of the collision domains to deploy.
+            excluded_links (Set[str]): A set containing the name of the collision domains to exclude.
 
         Returns:
             None
+
+        Raises:
+            InvocationError: If both `selected_links` and `excluded_links` are specified.
         """
-        links = {k: v for (k, v) in lab.links.items() if k in selected_links}.items() if selected_links \
-            else lab.links.items()
+        if selected_links and excluded_links:
+            raise InvocationError(f"You can either specify `selected_links` or `excluded_links`.")
+
+        links = lab.links.items()
+        if selected_links:
+            links = {
+                k: v for k, v in links if k in selected_links
+            }.items()
+        elif excluded_links:
+            links = {
+                k: v for k, v in links if k not in excluded_links
+            }.items()
 
         if len(links) > 0:
             pool_size = utils.get_pool_size()
-            link_pool = Pool(pool_size)
-
             items = utils.chunk_list(links, pool_size)
 
             EventDispatcher.get_instance().dispatch("links_deploy_started", items=links)
@@ -66,8 +78,9 @@ class KubernetesLink(object):
                     network_id: 1 for network_id in self._get_existing_network_ids()
                 })
 
-                for chunk in items:
-                    link_pool.map(func=partial(self._deploy_link, network_ids), iterable=chunk)
+                with Pool(pool_size) as links_pool:
+                    for chunk in items:
+                        links_pool.map(func=partial(self._deploy_link, network_ids), iterable=chunk)
 
             EventDispatcher.get_instance().dispatch("links_deploy_ended")
 
@@ -127,19 +140,18 @@ class KubernetesLink(object):
             None
         """
         networks = self.get_links_api_objects_by_filters(lab_hash=lab_hash)
-        if selected_links is not None:
+        if selected_links:
             networks = [item for item in networks if item["metadata"]["name"] in selected_links]
 
         if len(networks) > 0:
             pool_size = utils.get_pool_size()
-            links_pool = Pool(pool_size)
-
             items = utils.chunk_list(networks, pool_size)
 
             EventDispatcher.get_instance().dispatch("links_undeploy_started", items=networks)
 
-            for chunk in items:
-                links_pool.map(func=self._undeploy_link, iterable=chunk)
+            with Pool(pool_size) as links_pool:
+                for chunk in items:
+                    links_pool.map(func=self._undeploy_link, iterable=chunk)
 
             EventDispatcher.get_instance().dispatch("links_undeploy_ended")
 
@@ -152,12 +164,11 @@ class KubernetesLink(object):
         networks = self.get_links_api_objects_by_filters()
 
         pool_size = utils.get_pool_size()
-        links_pool = Pool(pool_size)
-
         items = utils.chunk_list(networks, pool_size)
 
-        for chunk in items:
-            links_pool.map(func=self._undeploy_link, iterable=chunk)
+        with Pool(pool_size) as links_pool:
+            for chunk in items:
+                links_pool.map(func=self._undeploy_link, iterable=chunk)
 
     def _undeploy_link(self, link_item: Any) -> None:
         """Undeploy a Kubernetes network.
@@ -197,7 +208,7 @@ class KubernetesLink(object):
         """
         filters = ["app=kathara"]
         if link_name:
-            filters.append("name=%s" % link_name)
+            filters.append(f"name={link_name}")
 
         # Get all Kathara namespaces if lab_hash is None
         namespaces = list(map(lambda x: x.metadata.name, self.kubernetes_namespace.get_all())) \
@@ -229,29 +240,35 @@ class KubernetesLink(object):
         Returns:
            Generator[Dict[str, KubernetesLinkStats], None, None]: A generator containing network names as keys and
                 KubernetesLinkStats as values.
-
-        Raises:
-            LinkNotFoundError: If the collision domains specified are not found.
         """
+        networks_stats = {}
+
+        def load_link_stats(network):
+            if network['metadata']['name'] not in networks_stats:
+                networks_stats[network['metadata']['name']] = KubernetesLinkStats(network)
+
         while True:
             networks = self.get_links_api_objects_by_filters(lab_hash=lab_hash, link_name=link_name)
             if not networks:
-                if not link_name:
-                    raise LinkNotFoundError("No collision domains found.")
-                else:
-                    raise LinkNotFoundError(f"Collision domains with name {link_name} not found.")
+                yield dict()
 
-            networks_stats = {}
+            pool_size = utils.get_pool_size()
+            items = utils.chunk_list(networks, pool_size)
+            with Pool(pool_size) as links_pool:
+                for chunk in items:
+                    links_pool.map(func=load_link_stats, iterable=chunk)
 
-            for network in networks:
-                networks_stats[network['metadata']['name']] = KubernetesLinkStats(network)
-
-            for network_stats in networks_stats.values():
+            networks_to_remove = []
+            for network_id, network_stats in networks_stats.items():
                 try:
                     network_stats.update()
                 except StopIteration:
+                    networks_to_remove.append(network_id)
                     continue
 
+            for k in networks_to_remove:
+                networks_stats.pop(k, None)
+
             yield networks_stats
 
     def _build_definition(self, link: Link, network_id: int) -> Dict[str, str]:
