@@ -1,13 +1,19 @@
 import hashlib
 import json
 import logging
+import os
 import re
 import shlex
+import signal
 import sys
+import tarfile
+import tempfile
+import threading
 import uuid
 from multiprocessing.dummy import Pool
-from typing import Optional, Set, List, Union, Generator, Tuple, Dict
+from typing import Optional, Set, List, Union, Generator, Tuple, Dict, Any
 
+import chardet
 from kubernetes import client
 from kubernetes.client.api import apps_v1_api
 from kubernetes.client.api import core_v1_api
@@ -17,19 +23,26 @@ from kubernetes.watch import watch
 
 from .KubernetesConfigMap import KubernetesConfigMap
 from .KubernetesNamespace import KubernetesNamespace
+from .exec_stream.KubernetesExecStream import KubernetesExecStream
 from .stats.KubernetesMachineStats import KubernetesMachineStats
 from ... import utils
 from ...event.EventDispatcher import EventDispatcher
-from ...exceptions import MachineAlreadyExistsError, MachineNotFoundError, MachineNotReadyError
+from ...exceptions import MachineAlreadyExistsError, MachineNotReadyError, MachineNotRunningError, MachineBinaryError, \
+    InvocationError
 from ...model.Lab import Lab
 from ...model.Machine import Machine
 from ...setting.Setting import Setting
 
 RP_FILTER_NAMESPACE = "net.ipv4.conf.%s.rp_filter"
 MAX_RESTART_COUNT = 3
+MAX_TIME_ERROR = 180
+
+OCI_RUNTIME_RE = re.compile(
+    r"OCI runtime exec failed"
+)
 
 # Known commands that each container should execute
-# Run order: shared.startup, machine.startup and machine.startup_commands
+# Run order: shared.startup, machine.startup and machine.meta['exec_commands']
 STARTUP_COMMANDS = [
     # If execution flag file is found, abort (this means that postStart has been called again)
     # If not flag the startup execution with a file
@@ -40,25 +53,28 @@ STARTUP_COMMANDS = [
     # Removes /etc/bind already existing configuration from k8s internal DNS
     "rm -Rf /etc/bind/*",
 
-    # Unmount the /etc/resolv.conf and /etc/hosts files, automatically mounted by Docker inside the container.
+    # Unmount the /etc/resolv.conf and /etc/hosts files, automatically mounted by Kubernetes inside the container.
     # In this way, they can be overwritten by custom user files.
     "umount /etc/resolv.conf",
     "umount /etc/hosts",
 
-    # Parse hostlab.b64
+    # Parse hostlab.b64 (if present)
+    "if [ -f \"/tmp/kathara/hostlab.b64\" ]; then "
     "base64 -d /tmp/kathara/hostlab.b64 > /hostlab.tar.gz",
     # Extract hostlab.tar.gz data into /
     "tar xmfz /hostlab.tar.gz -C /; rm -f hostlab.tar.gz",
+    "fi",
 
     # Copy the machine folder (if present) from the hostlab directory into the root folder of the container
     # In this way, files are all replaced in the container root folder
     "if [ -d \"/hostlab/{machine_name}\" ]; then "
-    "(cd /hostlab/{machine_name} && tar c .) | (cd / && tar xhf -); fi",
+    "(cd /hostlab/{machine_name} && tar c .) | (cd / && tar xhf - --no-same-owner --no-same-permissions); fi",
 
-    # If /etc/hosts is not configured by the user, add the localhost mapping
+    # If /etc/hosts is not configured by the user, add the default mappings
     "if [ ! -s \"/etc/hosts\" ]; then "
     "echo '127.0.0.1 localhost' > /etc/hosts",
     "echo '::1 localhost' >> /etc/hosts",
+    "echo '127.0.1.1 {machine_name}' >> /etc/hosts",
     "fi",
 
     # Give proper permissions to /var/www
@@ -67,12 +83,12 @@ STARTUP_COMMANDS = [
 
     # Give proper permissions to Quagga files (if present)
     "if [ -d \"/etc/quagga\" ]; then "
-    "chown quagga:quagga /etc/quagga/*",
+    "chown -R quagga:quagga /etc/quagga/",
     "chmod 640 /etc/quagga/*; fi",
 
     # Give proper permissions to FRR files (if present)
     "if [ -d \"/etc/frr\" ]; then "
-    "chown frr:frr /etc/frr/*",
+    "chown -R frr:frr /etc/frr/",
     "chmod 640 /etc/frr/*; fi",
 
     # If shared.startup file is present
@@ -94,10 +110,12 @@ STARTUP_COMMANDS = [
     "/hostlab/{machine_name}.startup &> /var/log/startup.log; fi",
 
     # Remove the Kubernetes' default gateway which points to the eth0 interface and causes problems sometimes.
-    "route del default dev eth0 || true",
+    "ip route del default dev eth0 || true",
 
     # Placeholder for user commands
-    "{machine_commands}"
+    "{machine_commands}",
+
+    "touch /tmp/EOS"
 ]
 
 SHUTDOWN_COMMANDS = [
@@ -125,43 +143,57 @@ class KubernetesMachine(object):
 
         self.kubernetes_namespace: KubernetesNamespace = kubernetes_namespace
 
-    def deploy_machines(self, lab: Lab, selected_machines: Set[str] = None) -> None:
+    def deploy_machines(self, lab: Lab, selected_machines: Set[str] = None, excluded_machines: Set[str] = None) -> None:
         """Deploy all the devices contained in lab.machines.
 
         Args:
             lab (Kathara.model.Lab.Lab): A Kathara network scenario.
             selected_machines (Set[str]): A set containing the name of the devices to deploy.
+            excluded_machines (Set[str]): A set containing the name of the devices to exclude.
 
         Returns:
             None
-        """
-        machines = {k: v for (k, v) in lab.machines.items() if k in selected_machines}.items() if selected_machines \
-            else lab.machines.items()
 
-        privileged = lab.general_options['privileged_machines'] if 'privileged_machines' in lab.general_options \
-            else False
-        if privileged:
-            logging.warning('Privileged option is not supported on Megalos. It will be ignored.')
+        Raises:
+            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
+        """
+        if selected_machines and excluded_machines:
+            raise InvocationError(f"You can either specify `selected_machines` or `excluded_machines`.")
+
+        machines = lab.machines.items()
+        if selected_machines:
+            machines = {
+                k: v for k, v in machines if k in selected_machines
+            }.items()
+        elif excluded_machines:
+            machines = {
+                k: v for k, v in machines if k not in excluded_machines
+            }.items()
 
         # Do not open terminals on Megalos
         Setting.get_instance().open_terminals = False
 
+        wait_thread = threading.Thread(
+            target=self._wait_machines_startup,
+            args=(lab, set([k for k, _ in machines]) if selected_machines or excluded_machines else None)
+        )
+        wait_thread.start()
+
         # Deploy all lab machines.
         # If there is no lab.dep file, machines can be deployed using multithreading.
         # If not, they're started sequentially
         if not lab.has_dependencies:
             pool_size = utils.get_pool_size()
-            machines_pool = Pool(pool_size)
-
             items = utils.chunk_list(machines, pool_size)
 
-            for chunk in items:
-                machines_pool.map(func=self._deploy_machine, iterable=chunk)
+            with Pool(pool_size) as machines_pool:
+                for chunk in items:
+                    machines_pool.map(func=self._deploy_machine, iterable=chunk)
         else:
             for item in machines:
                 self._deploy_machine(item)
 
-        self._wait_machines_startup(lab, selected_machines if selected_machines else None)
+        wait_thread.join()
 
     def _wait_machines_startup(self, lab: Lab, selected_machines: Set[str]) -> None:
         """Wait the startup of the selected machines. Return when the selected machines become `Ready`.
@@ -181,8 +213,27 @@ class KubernetesMachine(object):
         machines_ready = 0
         machines_failed = 0
 
+        # Create a timer to raise an exception if the execution gets stuck
+        def raise_timeout_error():
+            logging.error(
+                f"Network scenario startup is not responding for over {MAX_TIME_ERROR} seconds, exiting. "
+                f"To check devices status, use the following command:\n\t"
+                f"kubectl -n {lab.hash} get pods"
+            )
+
+            # We should send a SIGINT to the main thread to exit the w.stream
+            os.kill(os.getpid(), signal.SIGINT)
+
+        timer = threading.Timer(MAX_TIME_ERROR, raise_timeout_error)
+        timer.start()
+
         w = watch.Watch()
         for event in w.stream(self.kubernetes_namespace.client.list_namespaced_pod, namespace=lab.hash):
+            # Every new event, cancel and create the timer
+            timer.cancel()
+            timer = threading.Timer(MAX_TIME_ERROR, raise_timeout_error)
+            timer.start()
+
             machine_name = event['object'].metadata.labels['name']
 
             if not selected_machines or machine_name in selected_machines:
@@ -211,6 +262,9 @@ class KubernetesMachine(object):
                             logging.warning(f"Device `{machine_name}` has been restarted {restart_count} times.")
 
             if machines_ready + machines_failed == len(machines):
+                # Finished watching, cancel the last timer
+                timer.cancel()
+
                 w.stop()
 
         if machines_ready == len(machines):
@@ -243,16 +297,30 @@ class KubernetesMachine(object):
         """
         logging.debug("Creating device `%s`..." % machine.name)
 
-        # Get the general options into a local variable (just to avoid accessing the lab object every time)
-        options = machine.lab.general_options
+        # Get the global machine metadata into a local variable (just to avoid accessing the lab object every time)
+        global_machine_metadata = machine.lab.global_machine_metadata
+
+        # If privileged is defined for the device, throw a warning.
+        if "privileged" in global_machine_metadata or machine.is_privileged():
+            logging.warning(
+                f'Privileged option is not supported on Megalos. It will be ignored on device `{machine.name}`.'
+            )
 
         # If bridged is defined for the device, throw a warning.
-        if "bridged" in options or machine.meta['bridged']:
-            logging.warning('Bridged option is not supported on Megalos. It will be ignored.')
+        if "bridged" in global_machine_metadata or machine.is_bridged():
+            logging.warning(
+                f'Bridged option is not supported on Megalos. It will be ignored on device `{machine.name}`.'
+            )
+
+        # If any ulimit is defined fot the device throw a warning
+        if "ulimits" in machine.meta and len(machine.meta["ulimits"].keys()) > 0:
+            logging.warning(
+                f'Ulimit option is not supported on Megalos. It will be ignored on device `{machine.name}`.'
+            )
 
         # If any exec command is passed in command line, add it.
-        if "exec" in options:
-            machine.add_meta("exec", options["exec"])
+        if "exec" in global_machine_metadata:
+            machine.add_meta("exec", global_machine_metadata["exec"])
 
         # Sysctl params to pass to the container creation
         sysctl_parameters = {RP_FILTER_NAMESPACE % x: 0 for x in ["all", "default", "lo"]}
@@ -262,9 +330,15 @@ class KubernetesMachine(object):
 
         if machine.is_ipv6_enabled():
             sysctl_parameters["net.ipv6.conf.all.forwarding"] = 1
+            sysctl_parameters["net.ipv6.conf.all.accept_ra"] = 0
             sysctl_parameters["net.ipv6.icmp.ratelimit"] = 0
             sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 0
             sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 0
+        else:
+            sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 1
+            sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 1
+            sysctl_parameters["net.ipv6.conf.default.forwarding"] = 0
+            sysctl_parameters["net.ipv6.conf.all.forwarding"] = 0
 
         # Merge machine sysctls
         machine.meta['sysctls'] = {**sysctl_parameters, **machine.meta['sysctls']}
@@ -302,6 +376,22 @@ class KubernetesMachine(object):
         if Setting.get_instance().host_shared:
             volume_mounts.append(client.V1VolumeMount(name="shared", mount_path="/shared"))
 
+        # Mount volumes defined in the device
+        for idx, (host_path, volume) in enumerate(machine.meta["volumes"].items()):
+            missing_permissions = utils.check_directory_permissions(host_path, volume['mode'])
+
+            if not missing_permissions:
+                volume_mounts.append(
+                    client.V1VolumeMount(
+                        name=f"volume{idx}", mount_path=volume['guest_path'], read_only=volume['mode'] == "ro"
+                    )
+                )
+            else:
+                raise PermissionError(
+                    f"To mount volume `{host_path}` in `{volume['guest_path']}` "
+                    f"you miss the following permissions: `{', '.join(missing_permissions)}`."
+                )
+
         # Machine must be executed in privileged mode to run sysctls.
         security_context = client.V1SecurityContext(privileged=True)
 
@@ -338,11 +428,10 @@ class KubernetesMachine(object):
         # to execute custom commands coming from .startup file and "exec" option
         # Build the final startup commands string
         sysctl_commands = "; ".join(["sysctl -w -q %s=%d" % item for item in machine.meta["sysctls"].items()])
+        machine_commands = "; ".join(machine.meta['exec_commands']) if machine.meta['exec_commands'] else ":"
+
         startup_commands_string = "; ".join(STARTUP_COMMANDS) \
-            .format(machine_name=machine.name,
-                    sysctl_commands=sysctl_commands,
-                    machine_commands="; ".join(machine.startup_commands)
-                    )
+            .format(machine_name=machine.name, sysctl_commands=sysctl_commands, machine_commands=machine_commands)
 
         post_start = client.V1LifecycleHandler(
             _exec=client.V1ExecAction(
@@ -357,6 +446,11 @@ class KubernetesMachine(object):
         for env_var_name, env_var_value in machine.meta['envs'].items():
             env.append(client.V1EnvVar(env_var_name, env_var_value))
 
+        entrypoint = shlex.split(machine.meta["entrypoint"]) if "entrypoint" in machine.meta else None
+        args = machine.meta["args"] if "args" in machine.meta and machine.meta["args"] else None
+        if args:
+            args = shlex.split(args) if type(args) == str else args
+
         container_definition = client.V1Container(
             name=machine.meta['real_name'],
             image=machine.get_image(),
@@ -368,16 +462,23 @@ class KubernetesMachine(object):
             resources=resources,
             volume_mounts=volume_mounts,
             security_context=security_context,
-            env=env
+            env=env,
+            command=entrypoint,
+            args=args
         )
 
         pod_annotations = {}
         network_interfaces = []
-        for (idx, machine_link) in machine.interfaces.items():
+        for (idx, interface) in machine.interfaces.items():
+            additional_data = {"kathara.link": interface.link.name}
+            if interface.mac_address:
+                additional_data["mac"] = interface.mac_address
+
             network_interfaces.append({
-                "name": machine_link.api_object["metadata"]["name"],
+                "name": interface.link.api_object["metadata"]["name"],
                 "namespace": machine.lab.hash,
-                "interface": "net%d" % idx
+                "interface": "net%d" % idx,
+                **additional_data
             })
         pod_annotations["k8s.v1.cni.cncf.io/networks"] = json.dumps(network_interfaces)
 
@@ -415,12 +516,28 @@ class KubernetesMachine(object):
                 )
             ))
 
-        pod_spec = client.V1PodSpec(containers=[container_definition],
-                                    hostname=machine.meta['real_name'],
-                                    dns_policy="None",
-                                    dns_config=dns_config,
-                                    volumes=volumes
-                                    )
+        for idx, (host_path, volume) in enumerate(machine.meta["volumes"].items()):
+            volumes.append(client.V1Volume(
+                name=f"volume{idx}",
+                host_path=client.V1HostPathVolumeSource(
+                    path=host_path,
+                    type='DirectoryOrCreate'
+                )
+            ))
+
+        # Add Docker Config JSON for Private Registries
+        image_pull_secrets = []
+        if Setting.get_instance().docker_config_json:
+            image_pull_secrets.append(client.V1LocalObjectReference(name='private-registry'))
+
+        pod_spec = client.V1PodSpec(
+            containers=[container_definition],
+            hostname=machine.meta['real_name'],
+            dns_policy="None",
+            dns_config=dns_config,
+            volumes=volumes,
+            image_pull_secrets=image_pull_secrets
+        )
 
         pod_template = client.V1PodTemplateSpec(metadata=pod_metadata, spec=pod_spec)
         selector_rules = client.V1LabelSelector(match_labels=pod_labels)
@@ -436,34 +553,46 @@ class KubernetesMachine(object):
                                    spec=deployment_spec
                                    )
 
-    def undeploy(self, lab_hash: str, selected_machines: Optional[Set[str]] = None) -> None:
+    def undeploy(self, lab_hash: str, selected_machines: Set[str] = None, excluded_machines: Set[str] = None) -> None:
         """Undeploy all the running Kubernetes deployments and Pods contained in the scenario defined by the lab_hash.
 
         If selected_machines is not None, undeploy only the specified devices.
 
         Args:
             lab_hash (str): The hash of the network scenario to undeploy.
-            selected_machines (Optional[Set[str]]): If not None, undeploy only the specified devices.
+            selected_machines (Set[str]): A set containing the name of the devices to undeploy.
+            excluded_machines (Set[str]): A set containing the name of the devices to exclude.
 
         Returns:
             None
+
+        Raises:
+            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
         """
+        if selected_machines and excluded_machines:
+            raise InvocationError(f"You can either specify `selected_machines` or `excluded_machines`.")
+
         pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash)
-        if selected_machines is not None and len(selected_machines) > 0:
+        machines_to_watch = {item.metadata.labels["name"] for item in pods}
+        if selected_machines:
             pods = [item for item in pods if item.metadata.labels["name"] in selected_machines]
-        else:
-            selected_machines = {item.metadata.labels["name"] for item in pods}
+            machines_to_watch = selected_machines if selected_machines else machines_to_watch
+        elif excluded_machines:
+            pods = [item for item in pods if item.metadata.labels["name"] not in excluded_machines]
+            machines_to_watch = machines_to_watch if not excluded_machines else machines_to_watch - excluded_machines
 
         if len(pods) > 0:
-            pool_size = utils.get_pool_size()
-            machines_pool = Pool(pool_size)
+            wait_thread = threading.Thread(target=self._wait_machines_shutdown, args=(lab_hash, machines_to_watch))
+            wait_thread.start()
 
+            pool_size = utils.get_pool_size()
             items = utils.chunk_list(pods, pool_size)
 
-            for chunk in items:
-                machines_pool.map(func=self._undeploy_machine, iterable=chunk)
+            with Pool(pool_size) as machines_pool:
+                for chunk in items:
+                    machines_pool.map(func=self._undeploy_machine, iterable=chunk)
 
-            self._wait_machines_shutdown(lab_hash, selected_machines)
+            wait_thread.join()
 
     def _wait_machines_shutdown(self, lab_hash: str, selected_machines: Set[str]):
         """Wait the shutdown of the selected machines. Return when all the selected machines are terminated.
@@ -504,12 +633,11 @@ class KubernetesMachine(object):
         pods = self.get_machines_api_objects_by_filters()
 
         pool_size = utils.get_pool_size()
-        machines_pool = Pool(pool_size)
-
         items = utils.chunk_list(pods, pool_size)
 
-        for chunk in items:
-            machines_pool.map(func=self._undeploy_machine, iterable=chunk)
+        with Pool(pool_size) as machines_pool:
+            for chunk in items:
+                machines_pool.map(func=self._undeploy_machine, iterable=chunk)
 
     def _undeploy_machine(self, pod_api_object: client.V1Pod) -> None:
         """Undeploy a Kubernetes pod.
@@ -541,24 +669,15 @@ class KubernetesMachine(object):
         try:
             shell_env_value = self.get_env_var_value_from_pod(pod_api_object, "_MEGALOS_SHELL")
             shell = shell_env_value if shell_env_value else Setting.get_instance().device_shell
-            output = self.exec(machine_namespace,
-                               machine_name,
-                               command=[shell, '-c', shutdown_commands_string],
-                               )
-
-            try:
-                next(output)
-            except StopIteration:
-                pass
-
-            deployment_name = self.get_deployment_name(machine_name)
-            self.kubernetes_config_map.delete_for_machine(deployment_name, machine_namespace)
-
-            self.client.delete_namespaced_deployment(name=deployment_name,
-                                                     namespace=machine_namespace
-                                                     )
+            self.exec(machine_namespace, machine_name, command=[shell, '-c', shutdown_commands_string], is_stream=False)
         except ApiException:
-            return
+            pass
+        except MachineNotRunningError:
+            pass
+
+        deployment_name = self.get_deployment_name(machine_name)
+        self.kubernetes_config_map.delete_for_machine(deployment_name, machine_namespace)
+        self.client.delete_namespaced_deployment(name=deployment_name, namespace=machine_namespace)
 
     def connect(self, lab_hash: str, machine_name: str, shell: Union[str, List[str]] = None, logs: bool = False) \
             -> None:
@@ -574,11 +693,12 @@ class KubernetesMachine(object):
             None
 
         Raises:
+            MachineNotRunningError: If the specified device is not running.
             MachineNotReadyError: If the device is not ready.
         """
         pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)
         if not pods:
-            raise MachineNotFoundError("The specified device `%s` is not running." % machine_name)
+            raise MachineNotRunningError(machine_name)
         deployment = pods.pop()
 
         if 'Running' not in deployment.status.phase:
@@ -588,20 +708,23 @@ class KubernetesMachine(object):
             shell_env_value = self.get_env_var_value_from_pod(deployment, "_MEGALOS_SHELL")
             shell = shlex.split(shell_env_value if shell_env_value else Setting.get_instance().device_shell)
         else:
-            shell = shlex.split(shell) if type(shell) == str else shell
+            shell = shlex.split(shell) if type(shell) is str else shell
 
         logging.debug("Connect to device `%s` with shell: %s" % (machine_name, shell))
 
         if logs and Setting.get_instance().print_startup_log:
             exec_output = self.exec(lab_hash,
                                     machine_name,
-                                    command="/bin/cat /var/log/shared.log /var/log/startup.log"
+                                    command="/bin/cat /var/log/shared.log /var/log/startup.log /var/kathara/*"
                                     )
             try:
                 print("--- Startup Commands Log\n")
                 while True:
                     (stdout, _) = next(exec_output)
-                    stdout = stdout.decode('utf-8') if stdout else ""
+
+                    char_encoding = chardet.detect(stdout) if stdout else None
+
+                    stdout = stdout.decode(char_encoding['encoding']) if stdout else ""
                     sys.stdout.write(stdout)
             except StopIteration:
                 print("\n--- End Startup Commands Log\n")
@@ -652,25 +775,30 @@ class KubernetesMachine(object):
         return None
 
     def exec(self, lab_hash: str, machine_name: str, command: Union[str, List], tty: bool = False, stdin: bool = False,
-             stdin_buffer: List[Union[str, bytes]] = None, stderr: bool = False, is_stream: bool = False) \
-            -> Generator[Tuple[bytes, bytes], None, None]:
+             stdin_buffer: List[Union[str, bytes]] = None, stderr: bool = False, is_stream: bool = True) \
+            -> Union[KubernetesExecStream, Tuple[bytes, bytes, int]]:
         """Execute the command on the Kubernetes Pod specified by the lab_hash and the machine_name.
 
         Args:
             lab_hash (str): The hash of the network scenario containing the device.
             machine_name (str): The name of the device.
-            command (str): The command to execute.
+            command (Union[str, List]): The command to execute.
             tty (bool): If True, open a new tty.
             stdin (bool): If True, open the stdin channel.
             stdin_buffer (List[Union[str, bytes]]): List of command to pass to the stdin.
             stderr (bool): If True, return the stderr.
-            is_stream (bool): If True, return a generator with each line.
-                If False, returns a generator with the complete output.
+            is_stream (bool): If True, return a KubernetesExecStream object.
+                If False, returns a tuple containing the complete stdout, the stderr, and the return code of the command.
 
         Returns:
-            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
+             Union[KubernetesExecStream, Tuple[bytes, bytes, int]]: A KubernetesExecStream object or
+             a tuple containing the stdout, the stderr and the return code of the command.
+
+        Raises:
+            MachineNotRunningError: If the specified device is not running.
+            MachineBinaryError: If the command specified is not found on the device.
         """
-        command = shlex.split(command) if type(command) == str else command
+        command = shlex.split(command) if type(command) is str else command
 
         logging.debug("Executing command `%s` to device with name: %s" % (command, machine_name))
 
@@ -678,7 +806,7 @@ class KubernetesMachine(object):
             # Retrieve the pod of current Deployment
             pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)
             if not pods:
-                raise MachineNotFoundError("The specified device `%s` is not running." % machine_name)
+                raise MachineNotRunningError(machine_name)
             pod = pods.pop()
 
             response = stream(self.core_client.connect_get_namespaced_pod_exec,
@@ -691,41 +819,91 @@ class KubernetesMachine(object):
                               tty=tty,
                               _preload_content=False
                               )
-        except ApiException:
-            return None
+        except ApiException as e:
+            raise e
 
         if stdin_buffer is None:
             stdin_buffer = []
 
-        result = {
-            'stdout': '',
-            'stderr': ''
-        }
+        if is_stream:
+            exec_result = self._exec_stream(response, stdin, stdin_buffer, stderr)
+            return KubernetesExecStream(exec_result, response)
+        else:
+            return self._exec_all(response, machine_name, command, stdin, stdin_buffer, stderr)
 
+    @staticmethod
+    def _exec_stream(response: Any, stdin: bool = False, stdin_buffer: List[Union[str, bytes]] = None,
+                     has_stderr: bool = False) -> Generator[Tuple[bytes, bytes], None, None]:
+        """Execute the command on the Kubernetes Pod, returning a generator.
+
+        Args:
+            response (Any): The stream response from Kubernetes API.
+            stdin (bool): If True, open the stdin channel.
+            stdin_buffer (List[Union[str, bytes]]): List of command to pass to the stdin.
+            has_stderr (bool): If True, return the stderr.
+
+        Returns:
+            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
+        """
         while response.is_open():
             stdout = None
             stderr = None
             if response.peek_stdout():
                 stdout = response.read_stdout()
-                if not is_stream:
-                    result['stdout'] += stdout
-            if stderr and response.peek_stderr():
+            if has_stderr and response.peek_stderr():
                 stderr = response.read_stderr()
-                if not is_stream:
-                    result['stderr'] += stderr
             if stdin and stdin_buffer:
                 param = stdin_buffer.pop(0)
                 response.write_stdin(param)
                 if len(stdin_buffer) <= 0:
                     break
 
-            if is_stream and (stdout or stderr):
-                yield stdout.encode('utf-8') if stdout else None, stderr.encode('utf-8') if stderr else None
+            yield stdout.encode('utf-8') if stdout else None, stderr.encode('utf-8') if stderr else None
 
         response.close()
 
-        if not is_stream:
-            yield result['stdout'].encode('utf-8'), result['stderr'].encode('utf-8')
+    @staticmethod
+    def _exec_all(response: Any, machine_name: str, command: List, stdin: bool = False,
+                  stdin_buffer: List[Union[str, bytes]] = None, has_stderr: bool = False) -> Tuple[bytes, bytes, int]:
+        """Execute the command on the Kubernetes Pod, returning the full output.
+
+        Args:
+            response (Any): The stream response from Kubernetes API.
+            machine_name (str): The name of the device.
+            command (List): The command to execute.
+            stdin (bool): If True, open the stdin channel.
+            stdin_buffer (List[Union[str, bytes]]): List of command to pass to the stdin.
+            has_stderr (bool): If True, return the stderr.
+
+        Returns:
+            Tuple[bytes, bytes, int]: A tuple containing the stdout, the stderr and the return code of the command.
+
+        Raises:
+            MachineBinaryError: If the command specified is not found on the device.
+        """
+        result = {'stdout': '', 'stderr': ''}
+
+        while response.is_open():
+            if response.peek_stdout():
+                result['stdout'] += response.read_stdout()
+            if has_stderr and response.peek_stderr():
+                result['stderr'] += response.read_stderr()
+            if stdin and stdin_buffer:
+                param = stdin_buffer.pop(0)
+                response.write_stdin(param)
+                if len(stdin_buffer) <= 0:
+                    break
+
+        response.close()
+
+        try:
+            exit_code = response.returncode
+        except ValueError as e:
+            if OCI_RUNTIME_RE.search(str(e)):
+                raise MachineBinaryError(shlex.join(command), machine_name)
+            exit_code = 1
+
+        return result['stdout'].encode('utf-8'), result['stderr'].encode('utf-8'), exit_code
 
     def copy_files(self, machine_api_object: client.V1Deployment, path: str, tar_data: bytes) -> None:
         """Copy the files contained in tar_data in the Kubernetes deployment path specified by the machine_api_object.
@@ -753,6 +931,41 @@ class KubernetesMachine(object):
         except StopIteration:
             pass
 
+    def retrieve_files(self, machine_api_object: client.V1Deployment, src: str, dst: str) -> None:
+        """Get the file or directory from the Docker container specified by the machine_api_object into dst.
+
+        Args:
+            machine_api_object (docker.models.containers.Container): A Docker container.
+            src (str): The path of the file or folder to copy from the device.
+            dst (str): The destination path on the host.
+
+        Returns:
+            None
+        """
+        machine_name = machine_api_object.metadata.labels["name"]
+        machine_namespace = machine_api_object.metadata.namespace
+
+        # Get the tar from the Pod on the stdout
+        exec_output = self.exec(machine_namespace, machine_name, command=['tar', 'cf', '-', src], is_stream=True)
+
+        # Create a tmp tar file and write the output from the Pod
+        with tempfile.NamedTemporaryFile(mode='wb+', suffix='.tar') as temp_file:
+            # Write chunks in the temp tar file
+            while True:
+                try:
+                    stdout, _ = next(exec_output)
+                    if stdout:
+                        temp_file.write(stdout)
+                except StopIteration:
+                    break
+
+            # After writing, go at the beginning of file
+            temp_file.seek(0)
+
+            # Now, we need to extract the tmp tar file into the destination folder
+            with tarfile.open(fileobj=temp_file, mode='r') as tar_file:
+                tar_file.extractall(path=dst)
+
     def get_machines_api_objects_by_filters(self, lab_hash: str = None, machine_name: str = None) -> List[client.V1Pod]:
         """Return the List of Kubernetes Pods.
 
@@ -765,7 +978,7 @@ class KubernetesMachine(object):
         """
         filters = ["app=kathara"]
         if machine_name:
-            filters.append("name=%s" % machine_name)
+            filters.append(f"name={machine_name}")
 
         # Get all Kathara namespaces if lab_hash is None
         namespaces = list(map(lambda x: x.metadata.name, self.kubernetes_namespace.get_all())) \
@@ -794,25 +1007,34 @@ class KubernetesMachine(object):
             Generator[Dict[str, KubernetesMachineStats], None, None]: A generator containing device name as keys and
                 KubernetesMachineStats as values.
         """
+        machines_stats = {}
+
+        def load_machine_stats(pod):
+            if pod.metadata.name not in machines_stats:
+                machines_stats[pod.metadata.name] = KubernetesMachineStats(pod)
+
         while True:
             pods = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name)
             if not pods:
-                if not machine_name:
-                    raise MachineNotFoundError("No devices found.")
-                else:
-                    raise MachineNotFoundError(f"Devices with name {machine_name} not found.")
+                yield dict()
 
-            machines_stats = {}
+            pool_size = utils.get_pool_size()
+            items = utils.chunk_list(pods, pool_size)
+            with Pool(pool_size) as machines_pool:
+                for chunk in items:
+                    machines_pool.map(func=load_machine_stats, iterable=chunk)
 
-            for pod in pods:
-                machines_stats[pod.metadata.name] = KubernetesMachineStats(pod)
-
-            for machine_stats in machines_stats.values():
+            machines_to_remove = []
+            for machine_id, machine_stats in machines_stats.items():
                 try:
                     machine_stats.update()
                 except StopIteration:
+                    machines_to_remove.append(machine_id)
                     continue
 
+            for k in machines_to_remove:
+                machines_stats.pop(k, None)
+
             yield machines_stats
 
     @staticmethod
