@@ -1,27 +1,28 @@
+import math
+import sys
+from itertools import islice
+from multiprocessing import cpu_count
+from sys import platform as _platform
+
 import base64
 import hashlib
 import importlib
 import io
 import logging
-import math
 import os
 import re
 import shutil
-import sys
 import tarfile
 import tempfile
-from io import BytesIO
-from itertools import islice
-from multiprocessing import cpu_count
-from platform import node, machine
-from sys import platform as _platform
-from typing import Any, Optional, Match, Generator, List, Callable, Union, Dict, Iterable
-from types import ModuleType
 
 from binaryornot.check import is_binary
+from io import BytesIO
+from platform import node, machine
 from slug import slug
+from types import ModuleType
+from typing import Any, Optional, Match, Generator, List, Callable, Union, Dict, Iterable, Tuple
 
-from .exceptions import HostArchitectureError
+from .exceptions import HostArchitectureError, InvocationError
 
 # Platforms constants definition.
 MAC_OS: str = "darwin"
@@ -66,7 +67,7 @@ def get_executable_path(exec_path: str) -> Optional[str]:
     if os.path.exists(exec_abs_path) and os.path.isfile(exec_abs_path):
         # If kathara is launched as a python script
         exec_abs_path = "\"" + exec_abs_path + "\""
-        if exec_path.endswith(".py"):
+        if exec_path.endswith("kathara.py"):
             # Prepend python in windows because it has no shebang
             return exec_by_platform(lambda: exec_abs_path,
                                     lambda: "%s %s" % (sys.executable, exec_abs_path),
@@ -109,11 +110,34 @@ def get_pool_size() -> int:
     return cpu_count()
 
 
+def check_single_not_none_var(**kwargs) -> None:
+    not_none = [x for x in kwargs.values() if x is not None]
+
+    if len(not_none) > 1:
+        raise InvocationError(f"You must specify only a parameter among {', '.join(kwargs.keys())}")
+
+
+def check_required_single_not_none_var(**kwargs) -> None:
+    not_none = [x for x in kwargs.values() if x is not None]
+
+    if len(not_none) == 0:
+        raise InvocationError(f"You must specify a parameter among {', '.join(kwargs.keys())}")
+    elif len(not_none) > 1:
+        raise InvocationError(f"You must specify only a parameter among {', '.join(kwargs.keys())}")
+
+
 # Platform Specific Functions
 def is_platform(desired_platform: str) -> bool:
     return _platform == desired_platform
 
 
+def is_wsl_platform() -> bool:
+    if _platform == LINUX or _platform == LINUX2:
+        info = os.uname()
+        return 'microsoft' in info.release.lower()
+    return False
+
+
 def exec_by_platform(fun_linux: Callable, fun_windows: Callable, fun_mac: Callable) -> Any:
     if _platform == LINUX or _platform == LINUX2:
         return fun_linux()
@@ -141,33 +165,31 @@ def import_pywintypes() -> ModuleType:
     return exec_by_platform(pywintypes_import_stub, pywintypes_import_win, pywintypes_import_stub)
 
 
-# Architecture Test
-def get_architecture() -> str:
-    architecture = machine().lower()
-
-    logging.debug("Machine architecture is `%s`." % architecture)
-
-    if architecture in ["x86_64", "amd64"]:
-        return "amd64"
-    elif architecture == "i686":
-        return "386"
-    elif architecture in ["arm64", "aarch64"]:
-        return "arm64"
-    elif architecture == "armv7l":
-        return "armv7"
-    elif architecture == "armv6l":
-        return "armv6"
-    else:
-        raise HostArchitectureError(architecture)
+def wait_user_input_linux() -> list:
+    """Non-blocking input function for Linux and macOS."""
+    import select
+    to_break, _, _ = select.select([sys.stdin], [], [], 0.1)
+    return to_break
 
 
-def convert_win_2_linux(filename: str) -> bytes:
+def wait_user_input_windows() -> bool:
+    """Return True if an Enter keypress is waiting to be read. Only for Windows."""
+    import msvcrt
+    return b'\r' in msvcrt.getch() if msvcrt.kbhit() else False
+
+
+def convert_win_2_linux(filename: str, write: bool = False) -> Optional[bytes]:
     if not is_binary(filename):
         file_obj = None
         try:
             file_obj = open(filename, mode='r', encoding='utf-8-sig')
-            file_content = file_obj.read()
-            return file_content.replace("\n\r", "\n").encode('utf-8')
+            file_content = file_obj.read().replace("\n\r", "\n").replace("\r\n", "\n")
+            if not write:
+                return file_content.encode('utf-8')
+            else:
+                with open(filename, mode='w', encoding='utf-8', newline="\n") as file_obj_write:
+                    file_obj_write.write(file_content)
+                return
         except Exception:
             # In any case the binaryornot heuristics fail (so the file is not a text file), close the stream and
             # read the file as a standard binary file.
@@ -175,7 +197,8 @@ def convert_win_2_linux(filename: str) -> bytes:
                 file_obj.close()
             pass
 
-    return open(filename, mode='rb').read()
+    if not write:
+        return open(filename, mode='rb').read()
 
 
 def is_admin() -> bool:
@@ -241,6 +264,153 @@ def get_current_user_info() -> Any:
     return exec_by_platform(passwd_info, lambda: None, passwd_info)
 
 
+def check_directory_permissions(path: str, mode: str = "ro") -> list[str]:
+    """
+    Check directory permissions across different platforms.
+
+    Args:
+        path: Path to check permissions for
+        mode: Permission mode to check ('r' for read, 'w' for write, 'x' for execute)
+
+    Returns:
+        list[str]: A list containing any missing permissions
+    """
+
+    def check_directory_permissions_unix() -> list[str]:
+        if not os.path.exists(path):
+            raise FileExistsError(f"Path `{path}` does not exist.")
+
+        if not os.path.isdir(path):
+            raise NotADirectoryError(f"Path `{path}` must be a directory.")
+
+        missing = []
+
+        if 'r' in mode and not os.access(path, os.R_OK):
+            missing.append("read (r)")
+        if 'w' in mode and not os.access(path, os.W_OK):
+            missing.append("write (w)")
+        if 'x' in mode and not os.access(path, os.X_OK):
+            missing.append("execute (x)")
+
+        return missing
+
+    def check_directory_permissions_windows() -> list[str]:
+        import ctypes
+        from ctypes import wintypes
+
+        if not os.path.exists(path):
+            raise FileExistsError(f"Path `{path}` does not exist.")
+
+        if not os.path.isdir(path):
+            raise NotADirectoryError(f"Path `{path}` must be a directory.")
+
+        missing = []
+
+        generic_read = 0x80000000
+        generic_write = 0x40000000
+        generic_execute = 0x20000000
+
+        kernel32 = ctypes.WinDLL('kernel32', use_last_error=True)
+
+        create_file = kernel32.CreateFileW
+        create_file.argtypes = [
+            wintypes.LPCWSTR,
+            wintypes.DWORD,
+            wintypes.DWORD,
+            wintypes.LPVOID,
+            wintypes.DWORD,
+            wintypes.DWORD,
+            wintypes.HANDLE
+        ]
+        create_file.restype = wintypes.HANDLE
+
+        open_existing = 3
+        file_flag_backup_semantics = 0x02000000
+        invalid_handle_value = wintypes.HANDLE(-1).value
+        file_share_read = 0x00000001
+        file_share_write = 0x00000002
+
+        close_handle = kernel32.CloseHandle
+        close_handle.argtypes = [wintypes.HANDLE]
+        close_handle.restype = wintypes.BOOL
+
+        if 'r' in mode:
+            handle = create_file(
+                path,
+                generic_read,
+                file_share_read | file_share_write,
+                None,
+                open_existing,
+                file_flag_backup_semantics,
+                None
+            )
+
+            if handle == invalid_handle_value:
+                missing.append("read (r)")
+            else:
+                close_handle(handle)
+
+        if 'w' in mode:
+            handle = create_file(
+                path,
+                generic_write,
+                file_share_read | file_share_write,
+                None,
+                open_existing,
+                file_flag_backup_semantics,
+                None
+            )
+
+            if handle == invalid_handle_value:
+                missing.append("write (w)")
+            else:
+                close_handle(handle)
+
+        if 'x' in mode:
+            handle = create_file(
+                path,
+                generic_execute,
+                file_share_read | file_share_write,
+                None,
+                open_existing,
+                file_flag_backup_semantics,
+                None
+            )
+
+            if handle == invalid_handle_value:
+                missing.append("execute (x)")
+            else:
+                close_handle(handle)
+
+        return missing
+
+    return exec_by_platform(
+        check_directory_permissions_unix,
+        check_directory_permissions_windows,
+        check_directory_permissions_unix
+    )
+
+
+# Architecture Test
+def get_architecture() -> str:
+    architecture = machine().lower()
+
+    logging.debug("Machine architecture is `%s`." % architecture)
+
+    if architecture in ["x86_64", "amd64"]:
+        return "amd64"
+    elif architecture == "i686":
+        return "386"
+    elif architecture in ["arm64", "aarch64"]:
+        return "arm64"
+    elif architecture == "armv7l":
+        return "armv7"
+    elif architecture == "armv6l":
+        return "armv6"
+    else:
+        raise HostArchitectureError(architecture)
+
+
 # Formatting Functions
 def human_readable_bytes(size_bytes: int) -> str:
     if size_bytes == 0:
@@ -255,26 +425,6 @@ def human_readable_bytes(size_bytes: int) -> str:
 
 
 # Lab Functions
-def get_lab_temp_path(lab_name: str, force_creation: bool = True) -> str:
-    def windows_path():
-        import win32file
-        return win32file.GetLongPathName(tempfile.gettempdir())
-
-    tempdir = exec_by_platform(tempfile.gettempdir,
-                               windows_path,
-                               lambda: re.sub(r"/+", "/", "/%s" % get_absolute_path("/tmp"))
-                               )
-    lab_temp_directory = os.path.join(tempdir, lab_name)
-    if not os.path.isdir(lab_temp_directory) and force_creation:
-        os.mkdir(lab_temp_directory)
-
-    return lab_temp_directory
-
-
-def get_vlab_temp_path(force_creation: bool = True) -> str:
-    return get_lab_temp_path("kathara_vlab", force_creation=force_creation)
-
-
 def pack_file_for_tar(file_obj: Union[str, io.IOBase], arc_name: str) -> (tarfile.TarInfo, bytes):
     if isinstance(file_obj, str):
         file_content_patched = convert_win_2_linux(file_obj)
@@ -309,7 +459,38 @@ def pack_files_for_tar(guest_to_host: Dict) -> bytes:
     return tar_data
 
 
-def is_excluded_file(path: str) -> bool:
-    _, filename = os.path.split(path)
+def parse_cd_mac_address(value) -> Tuple[str, str]:
+    if '/' in value:
+        parts = [x for x in value.split('/') if x]
 
-    return filename in EXCLUDED_FILES
+        if len(parts) != 2:
+            raise SyntaxError(f"Invalid interface definition: `{value}`.")
+
+        cd_name = parts[0]
+        mac_address = parts[1]
+    else:
+        (cd_name, mac_address) = value, None
+
+    return cd_name, mac_address
+
+
+# Docker Engine Utils
+def parse_docker_engine_version(v: str) -> str:
+    parts = []
+    for part in v.split('.'):
+        numeric_part = ""
+        # Iterate over the single chars of the part
+        for char in part:
+            # As soon as we get a non-digit value, we break
+            if char.isdigit():
+                numeric_part += char
+            else:
+                break
+
+        # If the numeric_part is not empty, append it as a chunk of the version
+        if numeric_part:
+            parts.append(numeric_part)
+        else:
+            break
+
+    return '.'.join(parts)
